// GNU licence
// ROAM algoritmus pro generovani 3D krajiny
// Autor: Pavel ern (xcerny19)
// Vypracovno jako ronkov projekt na FIT.VUTBR

#include <Inventor/nodes/SoCoordinate3.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoEnvironment.h>
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoNormal.h>
#include <Inventor/nodes/SoNormalBinding.h> 
#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoIndexedTriangleStripSet.h>
#include <time.h>
#include <Inventor/sensors/SoFieldSensor.h>
#include <Inventor/nodes/SoCamera.h>
#include <vector>

#if defined(__WIN32__) || defined(_WIN32)
# include <Inventor/Win/viewers/SoWinExaminerViewer.h>
# pragma comment(linker,"/subsystem:\"console\"")    //Zapnuti konzole
#else
# include <Inventor/Qt/viewers/SoQtExaminerViewer.h>
#endif

using namespace std;

// --------------------------------------------------------------------------------------------
// --------------------- NASTAVOVANI PARAMETRU KRAJINY ----------------------------------------
// --------------------------------------------------------------------------------------------
// BEST NASTAVENI:
//	TERRAIN_DIMMENSION: 3, 5 ,9  ,17  ,33  ,65   ,129  , 257  ,513    ,1025
//  POSLEDNI_ID         14,62,254,1022,4094,16382,65534,262142,1048574,4194302
//  PRAH:				4, 6 ,30 ,30  ,60  ,120  ,230  , 400  ,
//	TERRAIN_HEIGHT:		1 ,2 ,2  ,10  ,10  ,30   ,80   , 120  ,
#define TERRAIN_DIMENSION 65    //(krajina 32x32 ctvercu se konstruuje ze 33x33 bod)
#define POSLEDNI_ID 16382       //4naN+1 - 2  ,  kde N je 2naN TERRAIN_DIMMENSION
#define PRAH_ZOBRAZENI 120 		//Udava jaka je hranice pomeru
#define PRAH_POSUNU_KAMERY 1    //Udava max zmenu polohy kamery kdy se nevola prepocet stromu
#define TERRAIN_HEIGHT_Q 30.f	// Maximln vska a hloubka krajiny
#define VZDALENOST_BODU 1		//Koeficient jak jsou od sebe vzdaleny body v krajine
#define POUZIJ_ROAM 			//Priznak zda pouzivat ROAM alg pro vykresleni
// --------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------
 

// Globln promnn a funkce,
// kter jsou pouity v modulech Main-*.cpp
const char *appTitle = "ROAM algoritmus";			// Titulek okna
bool headlight = false;//true;						// Defaultn svtlo (v pozici kamery)
SoSeparator* createScene(int argc, char **argv);	// Funkce vytvejc scnu
#if defined(__WIN32__) || defined(_WIN32)
SoWinExaminerViewer *viewer;
#else
SoQtExaminerViewer *viewer;
#endif
SoIndexedTriangleStripSet *strip;
SoCamera *camera;

struct trojuh{		//Struktura uzlu v BTT
  int IndBod1;			//1.bod trojuh
  int IndBod2;			//2.bod trojuh
  int IndBod3;			//3.bod trojuh
  bool Zobraz;			//Priznak zda je trojuh zobrazen  
  int Lsyn;				//ID leveho syna tohoto uzlu
  int Psyn;				//ID praveho syna tohoto uzlu
  int Otec;				//ID otce tohoto uzlu
};
struct bod_trojuh{  //Informace o zobrazenych bodech
  int  IDtrojuh [4];	//ID trojuh, ktery ma dany bod jako prostredni, -2=doplnovaci hodn
  bool Potvrzeni [4];   //info zda byl zobrazen
};

typedef int sezn_trojuh[(TERRAIN_DIMENSION-1)*(TERRAIN_DIMENSION-1)*2*4];

struct bod_trojuh Korekce[(TERRAIN_DIMENSION)*(TERRAIN_DIMENSION)]; //pole bodu
float hmap[TERRAIN_DIMENSION][TERRAIN_DIMENSION];	//Vyskova mapa
float RozMap[TERRAIN_DIMENSION][TERRAIN_DIMENSION];	//Rozdilova mapa - ma velikost jako mapa bodu
float Sour_bodu [(TERRAIN_DIMENSION)*(TERRAIN_DIMENSION)][3];//Kazdy bod z hmap ma tri souradnice XYZ
SbVec3f Bod_vector [(TERRAIN_DIMENSION)*(TERRAIN_DIMENSION)]; //Kazdy bod z hmap ma 1 sour urcujici vektor
trojuh Strom [POSLEDNI_ID+1];						//Zakladni struktura
SbTime t1,t2;										//Mereni casu



float randf(float max);
void generateMap();
void Vytvor_Novy_Strom();
void DalsiLevel(int &ID, trojuh OtecPomT,
				int Tr1BodX, int Tr1BodY, int Tr2BodX, int Tr2BodY, int Tr3BodX, int Tr3BodY,
				int IndBod1, int IndBod2, int IndBod3);


//------------------------------------------------------------------------------------------
// Funkce vracejc nhodnou float hodnotu v rozsahu -max a +max
float randf(float max)
{
  return (float(rand()) / float(RAND_MAX)) * max*2 - max;
}//fce

//------------------------------------------------------------------------------------------
//Funkce z tutorialu o OpenInventoru na serveru www.root.cz (autor: ing. Jan Peiva)

// Tato funkce vytvo metodou pesouvn prostednho bodu
// vkovou mapu, kter je vrcena funkc.
// Parametr dimension poet dk a sloupc vkov mapy
// a maxHeight maximln vku krajiny.
void generateMap(){	  // Pia poten vku do roh mapy
  hmap[0][0] = 0.f;
  hmap[0][TERRAIN_DIMENSION-1] = 0.f;
  hmap[TERRAIN_DIMENSION-1][0] = 0.f;
  hmap[TERRAIN_DIMENSION-1][TERRAIN_DIMENSION-1] = 0.f;

  // V nsledujc smyce se generuj postupn ni a ni detaily
  // mapy. Zan se s nejvtm krokem (promnn step nastavena na
  // velikost pole), a kon se s krokem 2, kdy jsou znmy vechny
  // body se sudou souadnic na obou osch a dopotvaj se
  // chybjc body mezi nimi.
  int step = TERRAIN_DIMENSION-1;
  float maxR = TERRAIN_HEIGHT_Q;
  int x,y;
  while (step >= 2) {

    // 1. fze: tvercov mka vodorovn
    // Ze ty bod tverce vygeneruj prostedn bod a k jeho hodnot
    // piti nhodn slo v rozsahu -maxR do +maxR
	  //Ja:   Vypocet prostrednich bodu ctverce  STR = (A+B+C+D)/4+RAND
    for (y=0; y<TERRAIN_DIMENSION-1; y+=step) {
      for (x=0; x<TERRAIN_DIMENSION-1; x+=step) {
        hmap[y+step/2][x+step/2] = (hmap[y][x] + hmap[y][x+step] +
            hmap[y+step][x] + hmap[y+step][x+step]) / 4.f + randf(maxR);
      }//for
    }//for

    // Korekce nhodnho rozptylu vky
    maxR *= (float)sqrt(0.5);

    // 2. fze: tvercov mka pootoen o 45 stup
    // Ze ty bod opt vygeneruj prostedn bod a jeho hodnotu
    // modifikuj. Po dokonen tto fze je hustota platnch
    // platnch dat v na vkov map dvojnsobn.	
    for (y=0; y<TERRAIN_DIMENSION-1; y+=step) {
      for (x=0; x<TERRAIN_DIMENSION-1; x+=step) {
        hmap[y][x+step/2] = ((y-step/2 >= 0 ? hmap[y-step/2][x+step/2] : 0) +
                            hmap[y][x] +
                            hmap[y][x+step] +
                            hmap[y+step/2][x+step/2]) /
                           (y-step/2 >= 0 ? 4 : 3) + randf(maxR);
        hmap[y+step/2][x] = (hmap[y][x] +
                            (x-step/2 >= 0 ? hmap[y+step/2][x-step/2] : 0) +
                            hmap[y+step/2][x+step/2] +
                            hmap[y+step][x]) /
                           (x-step/2 >= 0 ? 4 : 3) + randf(maxR);
        hmap[y+step/2][x+step] = (hmap[y][x+step] +
                                 hmap[y+step/2][x+step/2] +
                                 (x+step+step/2 < TERRAIN_DIMENSION ?
                                     hmap[y+step/2][x+step+step/2] : 0) +
                                 hmap[y+step][x+step]) /
                                (x+step+step/2 < TERRAIN_DIMENSION ? 4 : 3) +
                                randf(maxR);
        hmap[y+step][x+step/2] = (hmap[y+step/2][x+step/2] +
                                 hmap[y+step][x] +
                                 hmap[y+step][x+step] +
                                 (y+step+step/2 < TERRAIN_DIMENSION ?
                                     hmap[y+step+step/2][x+step/2] : 0)) /
                                (y+step+step/2 < TERRAIN_DIMENSION ? 4 : 3) +
                                randf(maxR);
      }//for
    }//for
    // Korekce nhodnho rozptylu vky
    maxR *= (float)sqrt(0.5);

    // Zmn krok na polovinu
    step /= 2;
  }//while
}//fce

//-------------------------------------------------------------------------------------------
// Funkce vypot normlu v danm bod vyskov mapy
SbVec3f calculateNormal(const int x, const int y){
float dx,dy;

  // hel v ose X
  if (x>=1 && x<=TERRAIN_DIMENSION-2)
    // Vpoet ze dvou sousednch bod
    dx = (hmap[x-1][y] - hmap[x+1][y]) / 2.f;
  else
    // Vpoet na okraji mapy
    if (x<1) dx = hmap[x][y] - hmap[x+1][y];
    else  dx = hmap[x-1][y] - hmap[x][y];
	
  // hel v ose Y
  if (y>=1 && y<=TERRAIN_DIMENSION-2)
    dy = (hmap[x][y-1] - hmap[x][y+1]) / 2.f;
  else
	if (y<1) dy = hmap[x][y] - hmap[x][y+1];
    else  dy = hmap[x][y-1] - hmap[x][y];

  SbVec3f r(dx, 1.f, dy);
  r.normalize();
  return r;
}//fce

//---------------------------------------------------------------------------------------------
//Umistuje pro kazdy bod na jadno z 4 mist(=max) info o IDtrojuh, ktery ma tento bod jako prostredni
void Umisteni_Korekce(int IndBod, int IDtr){	
  int ii = 0;

  for(; Korekce[IndBod].IDtrojuh[ii] != -2; ++ii);
  Korekce[IndBod].IDtrojuh[ii] = IDtr;  
}//fce
//-------------------------------------------------------------------------------------------
//Tato funkce vztvori BTT pro danou hmap
void Vytvor_Novy_Strom(){
    
  int ID = 0;		//Jsou cislovane podle poradi v tab = strome od 1
  trojuh PomT;      //Pomocny trojuhelnik
  int Tr1BodX;      //Musim si vzdy pamatovat tri vrcholy predchazejiciho trjuh abych vytvoril dalsi v naslednem levelu
  int Tr1BodY;		  //nelze provest predanim polem, protoze by se mi pri rekurzi prepsaly
  int Tr2BodX;
  int Tr2BodY;
  int Tr3BodX;
  int Tr3BodY;
  int IndBod1;
  int IndBod2;
  int IndBod3;

  //Vymazani pole korekce
  for (int p = 0; p < (TERRAIN_DIMENSION*TERRAIN_DIMENSION) ; ++p){    
    for (int ii = 0; ii < 4; ++ii){
      Korekce[p].IDtrojuh[ii] = -2;
	  Korekce[p].Potvrzeni[ii] = false;
	}//for
  }//for
  
  //Naplneni Sour_bodu
  int index = 0;
  for (int Ysour = 0; Ysour < (TERRAIN_DIMENSION); Ysour++){
    for (int Xsour = 0; Xsour < (TERRAIN_DIMENSION); Xsour++){
	  Sour_bodu[index][0] = Xsour * VZDALENOST_BODU;
	  Sour_bodu[index][1] = Ysour * VZDALENOST_BODU;
	  Sour_bodu[index][2] = hmap[Xsour][Ysour];
	  Bod_vector[index] = calculateNormal(Xsour, Ysour); 
      index++;
	}//for
  }//for

  //Pozn. mam pocet bodu = TERRAIN_DIMENSION, ale indexuju od nuly, takze rohy jsou 0 ... (TERRAIN_DIMENSION-1)
  ++ID;
  Tr1BodX = 0;                   Tr1BodY = 0;
  Tr2BodX = TERRAIN_DIMENSION-1; Tr2BodY = 0;
  Tr3BodX = TERRAIN_DIMENSION-1; Tr3BodY = TERRAIN_DIMENSION-1;
  PomT.Otec = 0;
   PomT.IndBod1 = Tr1BodX + Tr1BodY*TERRAIN_DIMENSION;
  IndBod1 = PomT.IndBod1;
   PomT.IndBod2 = Tr2BodX + Tr2BodY*TERRAIN_DIMENSION;
  Umisteni_Korekce(PomT.IndBod2, ID);
  IndBod2 = PomT.IndBod2;
   PomT.IndBod3 = Tr3BodX + Tr3BodY*TERRAIN_DIMENSION;
  IndBod3 = PomT.IndBod3;
  DalsiLevel(ID, PomT, Tr1BodX, Tr1BodY, Tr2BodX, Tr2BodY, Tr3BodX, Tr3BodY, IndBod1, IndBod2, IndBod3);

  ++ID;
  Tr2BodX = 0;                   Tr2BodY = TERRAIN_DIMENSION-1;
  PomT.Otec = 0;
  PomT.IndBod1 = IndBod1;
  PomT.IndBod2 = Tr2BodX + Tr2BodY*(TERRAIN_DIMENSION);
  IndBod2 = PomT.IndBod2;
  Umisteni_Korekce(PomT.IndBod2, ID);  
  PomT.IndBod3 = IndBod3;
  DalsiLevel(ID, PomT, Tr1BodX, Tr1BodY, Tr2BodX, Tr2BodY, Tr3BodX, Tr3BodY, IndBod1, IndBod2, IndBod3);  
  
  Strom[0].Lsyn = 1;
  Strom[0].Psyn = POSLEDNI_ID/2+1;  
  return ; 
}//fce
//--------------------------------------------------------------------------------------------
void DalsiLevel(int &ID, trojuh OtecPomT,
				int Tr1BodX, int Tr1BodY, int Tr2BodX, int Tr2BodY, int Tr3BodX, int Tr3BodY,
				int IndBod1, int IndBod2, int IndBod3){
    //Promene potrebne pro naplneni RozMap
  float Av;		//Prumerna vyska bodu (vypoc z krajnich bodu prepony)
  float Rozdil; //Rozdil dvou predchozich hodnot
    //Promene potrebne pro vytvoreni stromu
  trojuh PomT;
  int Novy_bodX, Novy_bodY;
  float Novy_bodZ;		//Sour noveho bodu
  int MeziVypX,MeziVypY;					//Mezivysledek pri vypoctu noveho bodu
  int PriznakPoslLevelu;//test zda jiz neni zadny bod na prepone trojuhelnika
  int IndNovy_bod;		//Index noveho bodu v poli indexu

  //Vypocet souradnic noveho bodu
  if (Tr1BodX == Tr3BodX){
    Novy_bodX = Tr1BodX;
  }else{
    MeziVypX = (abs(Tr1BodX - Tr3BodX)) / 2;
	if (Tr1BodX < Tr3BodX) Novy_bodX = Tr1BodX + MeziVypX;
	else Novy_bodX = Tr3BodX + MeziVypX;
  }
  if (Tr1BodY == Tr3BodY){
    Novy_bodY = Tr1BodY;
  }else{
    MeziVypY = (abs(Tr1BodY - Tr3BodY)) / 2;
	if (Tr1BodY < Tr3BodY) Novy_bodY = Tr1BodY + MeziVypY;
	else Novy_bodY = Tr3BodY + MeziVypY;
  }

  //Test zda nejsem na poslednim levelu
  if (MeziVypX < 1)
    if (MeziVypY < 1) PriznakPoslLevelu = 1;
    else PriznakPoslLevelu = 0;
  else PriznakPoslLevelu = 0;

  if (PriznakPoslLevelu == 0){
	IndNovy_bod = Novy_bodX + Novy_bodY*(TERRAIN_DIMENSION);
	PomT.IndBod2 = IndNovy_bod;
	Umisteni_Korekce(PomT.IndBod2, ID+1);
      
	//Plneni Rozdilove mapy - provede se vicekrat, vzdy pro novy bod - nevadi
    Novy_bodZ = hmap[Novy_bodX][Novy_bodY];
    Av = (hmap[Tr1BodX][Tr1BodY] + hmap[Tr3BodX][Tr3BodY]) / 2;
    Rozdil = fabsf( Novy_bodZ - Av);
    RozMap[Novy_bodX][Novy_bodY] = Rozdil;

	//Obstarani dalsiho uzlu (naplnim hodnoty a zanorim hloubeji, kde jeste doplnim a pak teprve ulozim)
    int Pam_ID_otce = ID;
    ++ID;
    PomT.Otec = Pam_ID_otce;
    OtecPomT.Lsyn = ID;
    PomT.IndBod1 = IndBod1;              //Sour 1.bodu = sour 1. bodu predchoziho trojuh
    PomT.IndBod3 = IndBod2;              //Sour 3.bodu = sour 2. bodu predchoziho trojuh (ale mezitim v poli dva body pribyly => -3+1-2=-4)
    DalsiLevel(ID, PomT, Tr1BodX, Tr1BodY, Novy_bodX, Novy_bodY, Tr2BodX, Tr2BodY, IndBod1, IndNovy_bod, IndBod2);

   ++ID;
    PomT.Otec = Pam_ID_otce;
    OtecPomT.Psyn = ID;
    PomT.IndBod1 = IndBod2;             //Sour 1.bodu = sour 2. bodu predchoziho trojuh
    PomT.IndBod2 = IndNovy_bod;         //Sour 2.bodu = nove vypocteny bod
	Umisteni_Korekce(PomT.IndBod2, ID);
    PomT.IndBod3 = IndBod3;             //Sour 3.bodu = sour 3. bodu predchoziho trojuh (ale mezitim v poli dva body pribyly => -3+1-2=-4)
    DalsiLevel(ID, PomT, Tr2BodX, Tr2BodY, Novy_bodX, Novy_bodY, Tr3BodX, Tr3BodY, IndBod2, IndNovy_bod, IndBod3);
    Strom[Pam_ID_otce] = OtecPomT;  //Ulozim otce do stromu

  }else {	//Jsem pod urovni nejmensiho trojuh (posledni level)= list ve strome
    OtecPomT.Lsyn = -1;			//Priznak ze jsem v listu
    OtecPomT.Psyn = -1;
    Strom[ID] = OtecPomT;
  }
}
//---------------------------------------------------------------------------------------------
void Vymazani_Korekce(){	
  for (int i = 0; i <= (TERRAIN_DIMENSION*TERRAIN_DIMENSION) ; ++i){    
	for (int ii = 0; ii < 4; ++ii){      
	  Korekce[i].Potvrzeni[ii] = false;
	}	
  }
}
//---------------------------------------------------------------------------------------------
void Potvrzeni_Korekce(int IndBod, int IDtr){	
  int ii = 0;
  for(; Korekce[IndBod].IDtrojuh[ii] != IDtr; ++ii);
  Korekce[IndBod].Potvrzeni[ii] = true;  
}//fce
//---------------------------------------------------------------------------------------------
void Vymazani_Priznaku_Zpracovani(){		//Zajistuje ze bratr, ktery je nasilne zobrazen svym bratrem uz nedostane priznak .zobraz = false
  for (int i = 1; i < (POSLEDNI_ID+1) ; ++i){    
	Strom[i].Zobraz = false;
  }
}
//---------------------------------------------------------------------------------------------
void Tisk_stromu(){		//Zajistuje ze bratr, ktery je nasilne zobrazen svym bratrem uz nedostane priznak .zobraz = false
  int PocID=0;
  bool Konec = false; 

  while(Konec == false){
    if (Strom[PocID].Lsyn != -1)  PocID = Strom[PocID].Lsyn;
	else{
	  if (Strom[ Strom[PocID].Otec ].Psyn != PocID) PocID = Strom[ Strom[PocID].Otec ].Psyn;
	  else{
	      PocID = Strom[PocID].Otec;
	      while((Strom[ Strom[PocID].Otec].Psyn == PocID)&&(Strom[PocID].Otec != 0)){
	        if (PocID == (POSLEDNI_ID/2+1+((POSLEDNI_ID/2+1)/2))) Konec = true; 
			PocID = Strom[PocID].Otec;
		  }//while           
		  PocID = Strom[ Strom[PocID].Otec ].Psyn;		  
	  }//else	
    }//else
	printf("ID: %d,Zobr:  %d\n",PocID,Strom[PocID].Zobraz);
  }//while
}
//---------------------------------------------------------------------------------------------
//funkce provede nastaveni priznaku zobrazeni vsech uzlu BTT
void Ohodnoceni (float Pozice_KameryX, float Pozice_KameryY, float Pozice_KameryZ){	//strom menim i ven
  int PocID = 1;
  float RozdilBodu;	//Chyba ktera by vznikla pri zaniku prostredniho bodu trojuhelniku(u praveho uhlu)
  float PomX, PomY, PomZ;	//Pom prom pro vypocet Vzdalenosti
  float Vzdalenost;	//Vzdalenost kamery od prostredniho bodu trojuhelniku(u praveho uhlu)
  bool Rozhodnuti;
  bool Konec = false;

  Strom[PocID++].Zobraz = true; //root trojuh vzdy
  Strom[POSLEDNI_ID/2+1].Zobraz = true; //root trojuh vzdy
  while (Konec == false){  
    //Nove vypoc bod(prostredni), rozhoduje jeho Hodnota z rozdilove mapy.
    float pom1=Sour_bodu[ Strom[PocID].IndBod2][0];	//x-sour prostredniho bodu trojuh s indexem PocID
    float pom2=Sour_bodu[ Strom[PocID].IndBod2][1];	//y-sour prostredniho bodu trojuh s indexem PocID
    float pom3=Sour_bodu[ Strom[PocID].IndBod2][2];	//z-sour prostredniho bodu trojuh s indexem PocID
    RozdilBodu = fabsf(RozMap [(int)(pom1 / VZDALENOST_BODU)][(int)(pom2 / VZDALENOST_BODU)]);			//z-sour prostredniho bodu trojuh s indexem PocID nezalezi na znamenku odchylky
    PomX = Pozice_KameryX - pom1;
    PomY = Pozice_KameryY - pom2;
    PomZ = Pozice_KameryZ - pom3;
    Vzdalenost = sqrt(PomX*PomX + PomY*PomY + PomZ*PomZ);
	Rozhodnuti = Vzdalenost/RozdilBodu < PRAH_ZOBRAZENI;
    
	if (Rozhodnuti){	
	  // nastav .zobraz true sebe i bratra	
      Strom[PocID].Zobraz = true;
	  Potvrzeni_Korekce(Strom[PocID].IndBod2, PocID);
      if (PocID == Strom[ Strom[PocID].Otec ].Lsyn){ //PocID je Lsyn sveho otce - zobraz i bratra
        int pom = Strom[ Strom[PocID].Otec ].Psyn;
	    Strom[pom].Zobraz = true;
		Potvrzeni_Korekce(Strom[pom].IndBod2, pom);
	  }
	  else{        
		int pom = Strom[ Strom[PocID].Otec ].Lsyn;
		Strom[pom].Zobraz = true;
        Potvrzeni_Korekce(Strom[pom].IndBod2, pom);
      }       	  
	  //dalsi uzel ve stromu kdyz je tento uzel .zobraz = true
      if (Strom[PocID].Lsyn != -1)  PocID = Strom[PocID].Lsyn;
	  else{
		if (Strom[ Strom[PocID].Otec ].Psyn != PocID) PocID = Strom[ Strom[PocID].Otec ].Psyn;
		PocID = Strom[PocID].Otec;
		while((Strom[ Strom[PocID].Otec].Psyn == PocID)&&(Strom[PocID].Otec != 0)){
          if (PocID == (POSLEDNI_ID/2+1+((POSLEDNI_ID/2+1)/2)))
		    Konec = true; 
		  PocID = Strom[PocID].Otec;
		}         
		PocID = Strom[ Strom[PocID].Otec ].Psyn;
		if (PocID == POSLEDNI_ID/2+1) PocID++;  //nesmi se stat ze by se nezobrazil druhy root trojuh
	  }//else dalsi uzel stromu 
    }else{
      //uz je .zobraz nastaveno na false inmlicitne, nebo true - zmeneno jeho bratrem = musi byt
      //dalsi uzel ve stromu kdyz je tento uzel .zobraz = false
      if (Strom[PocID].Otec != 0) PocID = Strom[PocID].Otec;      
	  while((Strom[ Strom[PocID].Otec].Psyn == PocID)&&(Strom[PocID].Otec != 0)){
	    PocID = Strom[PocID].Otec;
	  } 
	  if ((Strom[PocID].Otec == 0)&&(PocID == POSLEDNI_ID/2+1)) Konec = true; 
	  else{
		PocID = Strom[ Strom[PocID].Otec ].Psyn; 
		//ale tento uz je rozhodnuty ze sveho bratra tak jdu do jeho Lsyna
		if (Strom [PocID].Lsyn != -1) PocID = Strom [PocID].Lsyn;
	  }	  
	}//else rozhodnuti    
  }//while  
}//fce
//---------------------------------------------------------------------------------------------
void Naprava_otcu(int PocID){		
  int pom;
  int pamID;

  for (pom=0; pom<4; pom++){				//vsichni s bodem jako otec = true
    pamID = Korekce[ Strom[PocID].IndBod2 ].IDtrojuh[pom];
    if (pamID != -2){
	  Strom[pamID].Zobraz = true;					//nastavi v BTT                
	  Korekce[ Strom[PocID].IndBod2 ].Potvrzeni[pom] = true; 
      if (Strom[ Strom[pamID].Otec ].Zobraz == false) Naprava_otcu(Strom[pamID].Otec);
    }//if
  }//for				
}//fce
//---------------------------------------------------------------------------------------------
//funkce provede korekci priznaku zobrazeni tech uzlu BTT, ktere maji v poli korekce ruzne hodnoty 	Potvrzeni[]
void KorekceBTT (){
  bool actual;
  int pom;

  for (int i = 0; i < (TERRAIN_DIMENSION*TERRAIN_DIMENSION) ; ++i){        
	actual = Korekce[i].Potvrzeni[0];
	for (int ii = 0; ii < 4; ++ii){      
	  if ((Korekce[i].Potvrzeni[ii] != actual)&&(Korekce[i].IDtrojuh[ii] != -2)){//je chyba zobrazeni
		//byla zmena=>nektery trojuh zobrazeny v danem bodu, jiny ne => zobrazit vsechny
		for (int iii = 0; iii < 4; ++iii){
		  if ((Korekce[i].Potvrzeni[iii] == false)&&(Korekce[i].IDtrojuh[iii] != -2)){
		    Strom[ Korekce[i].IDtrojuh[iii] ].Zobraz = true;  //nastavim sebe v BTT pro zobraz = true
			Korekce[i].Potvrzeni[iii] = true;                 
			//vsichni otcove budou taky zobraz=true
			pom = Korekce[i].IDtrojuh[iii];
            if (Strom[ Strom[pom].Otec ].Zobraz == false) Naprava_otcu(Strom[pom].Otec);		 
          }//if
		}//for
		break;//tento uzel je vyrizeny
	  }//if
	}//for	
  }//for  
}//fce
//-------------------------------------------------------------------------------------------------
//
int Napln_Pole_Indexu(vector<int> &Pole_Indexu){

  int PocID = 1;			   //Ukazuje na kterem uzlu ve stromu se nachazim
  int PomPocID;
  int Velikost = 0;  
  bool Konec = false;

  while (Konec == false){
	while((Strom[PocID].Zobraz == true)&&(Strom[PocID].Lsyn != -1)){
      if (Strom[PocID].Lsyn != -1)  PocID = Strom[PocID].Lsyn;
	  else{
	    if (Strom[ Strom[PocID].Otec ].Psyn != PocID) PocID = Strom[ Strom[PocID].Otec ].Psyn;
	    else{
	      PocID = Strom[PocID].Otec;
	      while((Strom[ Strom[PocID].Otec].Psyn == PocID)&&(Strom[PocID].Otec != 0)){
	        if (PocID == (POSLEDNI_ID/2+1+((POSLEDNI_ID/2+1)/2))) Konec = true; 
			PocID = Strom[PocID].Otec;
		  }//while           
		  PocID = Strom[ Strom[PocID].Otec ].Psyn;		  
		}//else	
      }//else
	} //while
	if (Strom[PocID].Zobraz == false){
	  PomPocID = Strom[PocID].Otec;
	  PocID = Strom[PocID].Otec;
	  while((Strom[ Strom[PocID].Otec].Psyn == PocID)&&(Strom[PocID].Otec != 0)){
	    if (PocID == (POSLEDNI_ID/2+1+((POSLEDNI_ID/2+1)/2))) Konec = true; 
		PocID = Strom[PocID].Otec;
	  }//while           
	  if ((Strom[PocID].Otec == 0)&&(PocID == POSLEDNI_ID/2+1)) Konec = true;
	  PocID = Strom[ Strom[PocID].Otec ].Psyn;
    }//if
	else {
      PomPocID = PocID;
      if (Strom[ Strom[PocID].Otec ].Psyn != PocID) PocID = Strom[ Strom[PocID].Otec ].Psyn;
	  else{
	    PocID = Strom[PocID].Otec;
	    while((Strom[ Strom[PocID].Otec].Psyn == PocID)&&(Strom[PocID].Otec != 0)){
	      if (PocID == (POSLEDNI_ID/2+1+((POSLEDNI_ID/2+1)/2))) Konec = true; 
		  PocID = Strom[PocID].Otec;
		}//while           
	    PocID = Strom[ Strom[PocID].Otec ].Psyn;
	  }//else           			  
    }//else
	Velikost++;
    Pole_Indexu.push_back(Strom[PomPocID].IndBod1);
	Pole_Indexu.push_back(Strom[PomPocID].IndBod2);
	Pole_Indexu.push_back(Strom[PomPocID].IndBod3);
	Pole_Indexu.push_back(-1);
  }//while
  return Velikost;
}//fce

//----------------------------------------------------------------------------------------------
// Funkce vytvo model nas vskov mapy
float VektorZmenyPozice(float PosKamX, float PosKamY, float PosKamZ,
						float PosMinulX, float PosMinulY, float PosMinulZ){  
  float PomX, PomY, PomZ;
  float Vysledek = 0;

  PomX = PosKamX - PosMinulX;
  PomY = PosKamY - PosMinulY;
  PomZ = PosKamZ - PosMinulZ;
  Vysledek = sqrt(PomX*PomX + PomY*PomY + PomZ*PomZ);  
  return Vysledek;
}//fce

//----------------------------------------------------------------------------------
void cameraPositionChanged(void *data, SoSensor * sensor){
  vector<int> Pole_Indexu;
  int Velikost_pole;
  static float PosMinuleX = 200;
  static float PosMinuleY = 200;
  static float PosMinuleZ = 200;
  
  const SbVec3f &pos = camera->position.getValue();  
  if (VektorZmenyPozice(pos[0], pos[1], pos[2], PosMinuleX, PosMinuleY, PosMinuleZ)
	     > PRAH_POSUNU_KAMERY){ 
    //Doslo k prepocitani BTT, provedu ulozeni nove pozice pro nezmenitelny strom
	PosMinuleX = pos[0];
	PosMinuleY = pos[1];
	PosMinuleZ = pos[2];
    //Pokud se kamera posune o vetsi kus tak ohodnot znovu strom
	trojuh *Strom = (trojuh*)data;    
    Vymazani_Korekce();
    //Mereni casu
      t2 = SbTime::getTimeOfDay();
      printf("Vymazani korekce: %f\n-----------------------\n",t2-t1);
      t1 = SbTime::getTimeOfDay();

    Vymazani_Priznaku_Zpracovani();
    //Mereni casu
      t2 = SbTime::getTimeOfDay();
      printf("Vymazani priznaku zprac: %f\n-----------------------\n",t2-t1);
      t1 = SbTime::getTimeOfDay();

    Ohodnoceni (pos[0], pos[1], pos[2]);
    //Mereni casu
      t2 = SbTime::getTimeOfDay();
      printf("Ohodnoceni:       %f\n-----------------------\n",t2-t1);
      t1 = SbTime::getTimeOfDay();

    KorekceBTT ();
    //Mereni casu
      t2 = SbTime::getTimeOfDay();
      printf("Korekce BTT:      %f\n-----------------------\n",t2-t1);
      t1 = SbTime::getTimeOfDay();

    Velikost_pole = Napln_Pole_Indexu(Pole_Indexu);  
    //Mereni casu
      t2 = SbTime::getTimeOfDay();
      printf("Napln pole indexu:%f\n\n\n",t2-t1);
      t1 = SbTime::getTimeOfDay();

    int* PoleIndices = new int[Velikost_pole*4];
    for(int pom=0; pom < Velikost_pole*4; pom++) PoleIndices[pom] = Pole_Indexu[pom];	  
    strip->coordIndex.setNum(Velikost_pole*4);
    strip->coordIndex.setValues(0, (Velikost_pole * 4), PoleIndices);
  }//if
}//fce

//----------------------------------------------------------------------------------
// Funkce vytvo graf scny pro nasi aplikaci.
SoSeparator* createScene(int argc, char **argv)
{
  // Inicializace genertoru nhodnch sel
  srand((unsigned int)time(NULL));

  // Vytvo koen scny
  SoSeparator *root = new SoSeparator;
  //root->renderCaching = SoSeparator::OFF;
  root->ref();
  viewer->setSceneGraph(root);

  // Nastav intenzitu globlnho svtla ve scn
  SoEnvironment *env = new SoEnvironment;
  env->ambientIntensity.setValue(0.5);
  root->addChild(env);

  // Svtlo
  SoDirectionalLight *sunLight = new SoDirectionalLight;
  sunLight->direction.setValue(TERRAIN_DIMENSION/2, -TERRAIN_HEIGHT_Q*4, TERRAIN_DIMENSION/2);
  sunLight->intensity.setValue(1.0f);
  root->addChild(sunLight);

  // Vytvori vyskovou mapu
  generateMap();
// -----------------------------Pouziti ROAM algoritmu------------------------------------
#ifdef POUZIJ_ROAM
  //Vytvori BTT
  trojuh *Strom;
  //Mereni casu
    t1 = SbTime::getTimeOfDay();    
  Vytvor_Novy_Strom();
  //Mereni casu
    t2 = SbTime::getTimeOfDay();  
    printf("Vytvor_strom:     %f\n-----------------------\n",t2-t1);
    t1 = SbTime::getTimeOfDay();  

  //Souradnice bodu renderovane mapy
  SoCoordinate3 *coords = new SoCoordinate3;
  coords->point.setValues(0, (TERRAIN_DIMENSION)*(TERRAIN_DIMENSION), Sour_bodu);
  root->addChild(coords);

  //Normaly
  // Pro normly nastavujeme binding, tj. zda-li specifikujeme
  // normly pro kad bod (PER_VERTEX), nebo jen pro trojhelnky
  // (PER_FACE), i cel plocha m jen jednu normlu (OVERALL).
  SoNormal *normals = new SoNormal; 
  SoNormalBinding *nbind = new SoNormalBinding;
  nbind->value.setValue(SoNormalBinding::PER_VERTEX_INDEXED);
  root->addChild(nbind); 
  normals->vector.setValues(0, (TERRAIN_DIMENSION)*(TERRAIN_DIMENSION), Bod_vector);
  root->addChild(normals); 

  // registrace callbacku zmeny pozice kamery
  camera = viewer->getCamera();
  SoFieldSensor *sensor = new SoFieldSensor(cameraPositionChanged, &Strom);
  sensor->attach(&camera->position);

  // Materil (barva) krajiny
  SoMaterial *material = new SoMaterial;
  material->ambientColor.setValue(0.0f, 0.0f, 0.0f);
  material->diffuseColor.setValue(0.0f, 0.9f, 0.0f);
  root->addChild(material);

  strip = new SoIndexedTriangleStripSet;
  root->addChild(strip);
  cameraPositionChanged(Strom, sensor);
 
// -----------------------------Bez ROAM algoritmu------------------------------------
#else
   // Materil (barva) krajiny
  SoMaterial *material = new SoMaterial;
  material->ambientColor.setValue(0.0f, 0.0f, 0.0f);
  material->diffuseColor.setValue(0.0f, 0.9f, 0.0f);
  root->addChild(material);
   //Naplneni poli coord pr vykresleni
  static float Body[TERRAIN_DIMENSION*TERRAIN_DIMENSION][3];
  int poc = -1;
  int pocx, pocy;
  for(pocy=0; pocy<TERRAIN_DIMENSION; ++pocy){
    for(pocx=0; pocx<TERRAIN_DIMENSION; ++pocx){
       Body[++poc][0] = pocx;//pocx;
	   Body[poc]  [1] = pocy;//hmap[pocx][pocy];
	   Body[poc]  [2] = hmap[pocx][pocy];//pocy;
	}//for
  }//for
  static int32_t BodyIndices[(TERRAIN_DIMENSION-1)*(TERRAIN_DIMENSION-1)][5];
  poc = -1;
  for(pocy=0; pocy<TERRAIN_DIMENSION-1; ++pocy){
    for(pocx=0; pocx<TERRAIN_DIMENSION-1; ++pocx){
       BodyIndices[++poc][0] = pocy * TERRAIN_DIMENSION + pocx;	  
	   BodyIndices[poc]  [1] = BodyIndices[poc][0]+TERRAIN_DIMENSION;
	   BodyIndices[poc]  [2] = BodyIndices[poc][0]+1;	  
	   BodyIndices[poc]  [3] = BodyIndices[poc][0]+TERRAIN_DIMENSION+1;
	   BodyIndices[poc]  [4] = -1;	   
	}//for
  }//for
  SoCoordinate3 *coords = new SoCoordinate3;
  coords->point.setValues(0, TERRAIN_DIMENSION*TERRAIN_DIMENSION, Body);
  root->addChild(coords); 

  SoIndexedTriangleStripSet *strip = new SoIndexedTriangleStripSet;
  strip->coordIndex.setValues(0, (poc+1) * 5, (int*)&BodyIndices);   
  root->addChild(strip);
#endif  //else bez ROAM
  root->unrefNoDelete();
  return root;
}

